import torch
import torch.nn as nn
import torch.optim as optim

from scratchai.learners import metrics
from tabulate import tabulate
from tqdm import tqdm


__all__ = ['Learner', 'SegLearner']

  
class Learner():
  
  def __init__(self, net:nn.Module, loader=None, lr:float=1e-3, 
        epochs:int=10, mt:str='seg', sae:int=1, pe:int=1, wd:int=1e-4,
        trainiter:int=None, valiter:int=None, 
        crit:nn.Module=nn.CrossEntropyLoss, opt=optim.Adam, she:int=5,
        device='cuda', metrics:list=None):
    """
    A Learner Object. Base Class for all Learners.

    Arguments
    ---------
    net : nn.Module
       The model to train
    loader : TOFILL
        The DataLoader from where to get the data
    epochs : int
        The number of epochs
    she : int
       Interval in which to checkpoint
    pe : int
      Interval in which it prints the loss and other metrics
    wd : float
      Weight Decay for L2 Regularization

    """

    self.lr = lr
    self.epochs = epochs
    self.mt = mt
    self.device = device
    self.net = net.to(self.device)
    self.she = she
    self.sae = sae
    self.pe = pe
    self.loader = loader
    self.metrics = metrics
    self.wd = wd
    self.trainiter = trainiter
    self.valiter = valiter

    self.tlosses = []
    self.vlosses = []

    '''
    self.dltr = self.loader.get_dltr()
    self.dltt = self.loader.get_dltt()
    self.trainiter = len(self.dltr) // self.dltr.batch_size if not trainiter \
            else trainiter
    self.valiter = len(self.dltt) // self.dltt.batch_size if not valiter \
            else valiter
    '''
    self.crit = crit()
    self.opt = opt(net.parameters(), lr=self.lr, weight_decay=self.wd)

    self.h = 512
    self.w = 512
  
  def count_params(self):
      return sum(p.numel() for p in self.net.parameters() if p.requires_grad)

  def fit(self):
    pass
  
  def calc_metrics(self):
    pass

  ################################################################################
  ######################## Ongoing Print Functions ###############################
  ################################################################################
  # TODO Use the torch-summary module and build on that.

  def conv_out_size(self, net):
    kh, kw = net.kernel_size if type(net.kernel_size) == tuple else (net.kernel_size, net.kernel_size)
    sh, sw = net.stride if type(net.stride) == tuple else (net.stride, net.stride)
    ph, pw = net.padding if type(net.padding) == tuple else (net.padding, net.padding)

    self.h = (int) ((self.h - kh + (2 * ph)) / sh) + 1
    self.w = (int) ((self.w - kw + (2 * pw)) / sw) + 1
    return self.h, self.w

  def unet_eblock_out(self, net):
    self.h = (int) ((self.h * 2) - 4)
    self.w = (int) ((self.w * 2) - 4)
    return self.h, self.w
  
  def summary(self):
    layers = [['Input'], [(self.h, self.w)]]
    print (tabulate(layers))
    self._summary(self.net)

  def _summary(self, net):
    layers = []
    for m in net.children():
      temp = []
      if isinstance(m, nn.Sequential):
        self._summary(m)

      elif isinstance(m, nn.Conv2d):
        temp.append('Conv2d({}, {}, {})'.format(m.kernel_size, m.stride, m.padding))
        temp.append('{}'.format(self.conv_out_size(m)))

      elif str(m.__class__).split('.')[-1][:-2] == 'UNet_EBlock':
        temp.append('UNet_EBlock({}, {})'.format(m.uc.in_channels, m.uc.out_channels))
        temp.append('{}'.format(self.unet_eblock_out(m)))

      elif str(m.__class__).split('.')[-1][:-2] == 'MaxPool2d':
        temp.append('MaxPool2d({}, {}, {})'.format(m.kernel_size, m.stride, m.padding))
        temp.append('{}'.format(self.conv_out_size(m)))
      else:
        temp.append('ReLU')
        temp.append('{}'.format((self.h, self.w)))
      
      if len(temp) > 0:
        layers.append(temp)
    
    print (tabulate(layers))


class SegLearner(Learner):
  """
  The Learner Object that helps train Segmentation Datasets.
  """
  
  # TODO Add pixel accuracy metric function to this
  avbl_metrics = ['miou']

  def __init__(self, *args, **kwargs):
    
    super().__init__(*args, **kwargs)
    
    '''
    self.dltr = self.loader.get_dltr()
    self.dltt = self.loader.get_dltt()
    '''
    self.trainiter = self.loader.len if not self.trainiter \
            else self.trainiter
    self.valiter = self.loader.len if not self.valiter \
            else self.valiter
    
    # Check if no extra metrics are requested than supported and 
    if self.metrics:
      assert len(SegLearner.avbl_metrics) >= len(self.metrics)
      assert all([True if metric in self.avbl_metrics else False \
                  for metric in self.metrics])

  def fit(self):
    """
    Method to help in training of Segmentation datasets
    """

    for e in range(1, self.epochs+1):
      trloss = 0
      self.net.train()

      for ii in tqdm(range(self.trainiter)):

        img, lab = next(iter(self.loader.get_batch()))

        # TODO hack to skip batches with size = 1
        if img.shape[0] == 1:
          continue

        img, lab = img.to(self.device), lab.to(self.device)
        # TODO Introduce transforms in get_batch
        #lab *= 255

        self.opt.zero_grad()
        out = self.net(img.float())
        loss = self.crit(out, lab.long())
        loss.backward()
        self.opt.step()

        trloss += loss.item()

        self.calc_metrics(out, lab)
        break

      self.tlosses.append(trloss)

      if ii % self.she == 0:
        #show_camvid(self.net, self.h, self.w)
        print ('Epoch {}/{}...'.format(e, self.epochs),
          'Loss {:3f}'.format(loss.item()))

      if e % self.sae == 0:
        checkpoint = {
          'epochs' : e,
          'model_state_dict' : self.net.state_dict(),
          'opt_state_dict' : self.opt.state_dict()
        }
        torch.save(checkpoint, './ckpt-{:2f}.pth'.format(trloss))
        print ('Model saved!')
        
      if e % self.pe == 0:
        print ('Epoch {}/{}...'.format(e, self.epochs),
            'Loss {:3f}'.format(trloss))

      if e % eval_every == 0 and self.dltt:
        with torch.no_grad():
          self.net.eval()

          eloss = 0

          for _ in tqdm(range(self.valiter)):
            img, lab = next(iter(self.get_batch()))

            img, lab = img.to(device), lab.to(device)
            out = self.net(img.float())
            loss = criterion(out, lab.long())
            eloss += loss.item()

          print ('Loss {:3f}'.format(eloss // self.valiter))

          self.vlosses.append(eloss)
      
      # TODO: An API where users can easily setup a scheduler
      #scheduler.step(train_loss)

    def calc_metrics(self, inp, lab):
      # TODO Implement
      return
      for metric in metrics:
        setattr(self, metric, getattr(metrics, metric)(inp, lab))



import unittest
import scratchai

class TestLoaders(unittest.TestCase):

    def test_seg(self):
        l = scratchai.SegLoader('./vault/camvid/images', './vault/camvid/labels', 2)
        self.assertEqual(type(l), scratchai.DataLoader.SegLoader.SegLoader, 'Type Mismatch')



"""
The Tests that needs to be performed on scratchai.nets
"""

import sys
import os
import unittest
import torch
import torch.nn as nn
import numpy as np

import scratchai
from scratchai import *

class TestUNet(unittest.TestCase):
  def test_paper(self):
    '''
    This module tests the UNet implementation to be the same
    as shown in the paper.
    '''
    noise = torch.randn(2, 3, 572, 572)
    # TODO Assert device is on cpu after initialization

    unet1 = scratchai.nets.UNet(3, 4, sos=False)
    out = unet1(noise)
    self.assertEqual(list(out.shape), [2, 4, 388, 388], "The out shape not same as in shape")

    unet2 = scratchai.nets.UNet(3, 4, sos=True)
    out = unet2(noise)
    self.assertEqual(list(out.shape), [2, 4, 572, 572], "The out shape not same as in shape")

  def test_conv(self):
    conv = scratchai.nets.seg.unet.conv(3, 3)[0]
    noise = torch.randn(2, 3, 4, 4)
    out = conv(noise)
    self.assertEqual(list(out.shape), [2, 3, 2, 2], "The out shape not same as in shape")

  def test_uconv(self):
    conv = scratchai.nets.seg.unet.uconv(3, 3)[0]
    noise = torch.randn(2, 3, 2, 2)
    out = conv(noise)
    self.assertEqual(list(out.shape), [2, 3, 4, 4], "The out shape not same as in shape")

  def test_ublock(self):
    '''
    Input: [N, C, H, W]
    Output: [N, C//2, (H*2) - 4, (W*2) - 4]
    '''
    net = scratchai.nets.seg.unet.UNet_EBlock(4)
    n1 = torch.randn(2, 4, 52, 52)
    n2 = torch.randn(2, 2, 136, 136)
    out = net(n1, n2)
    self.assertEqual(list(out.shape), [2, 2, 100, 100], "The out shape not same as in shape")

class TestENet(unittest.TestCase):

  def test_initial_block(self):
    noise = torch.randn(2, 3, 4, 4)
    net = scratchai.nets.seg.enet.InitialBlock(3, 6)
    out = net(noise)
    self.assertEqual(list(out.shape), [2, 6, 2, 2], "out shape reduction not as it should"
                              " be.")
  def test_RDANeck(self):
    noise = torch.randn(2, 8, 4, 4)
    
    # Check 1
    net = scratchai.nets.seg.enet.RDANeck(8, 8, device='cpu')
    out = net(noise)
    self.assertEqual(list(out.shape), [2, 8, 4, 4], "out shape reduction not as it should"
                              " be.")
    # Check 2
    net = scratchai.nets.seg.enet.RDANeck(8, 9, device='cpu')
    out = net(noise)
    self.assertEqual(list(out.shape), [2, 9, 4, 4], "out shape reduction not as it should"
                              " be.")
    # Check 3
    net = scratchai.nets.seg.enet.RDANeck(8, 8, aflag=True, device='cpu')
    out = net(noise)
    self.assertEqual(list(out.shape), [2, 8, 4, 4], "out shape reduction not as it should"
                              " be.")
  def test_DNeck_UNeck(self):
    noise = torch.randn(2, 8, 4, 4)
    noise2 = torch.randn(2, 8, 2, 2)
    
    # Check 1
    net = scratchai.nets.seg.enet.DNeck(8, 8, device='cpu')
    out, idxs = net(noise)
    self.assertEqual(list(out.shape), [2, 8, 2, 2], "out shape reduction not as it should"
                              " be.")
    net = scratchai.nets.seg.enet.UNeck(8, 8)
    out = net(noise2, idxs)
    self.assertEqual(list(out.shape), [2, 8, 4, 4], "out shape reduction not as it should"
                              " be.")
    
    # Check 2
    net = scratchai.nets.seg.enet.DNeck(8, 9, device='cpu')
    out, idxs = net(noise)
    self.assertEqual(list(out.shape), [2, 9, 2, 2], "out shape reduction not as it should"
                              " be.")
    # Check 3
    noise = torch.randn(2, 7, 4, 4)
    noise2 = torch.randn(2, 8, 2, 2)

    net = scratchai.nets.seg.enet.DNeck(7, 7, device='cpu')
    out, idxs = net(noise)
    self.assertEqual(list(out.shape), [2, 7, 2, 2], "out shape reduction not as it should"
                              " be.")
    net = scratchai.nets.seg.enet.UNeck(8, 7)
    out = net(noise2, idxs)
    self.assertEqual(list(out.shape), [2, 7, 4, 4], "out shape reduction not as it should"
                             " be.")
  
  def test_enet(self):
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    n1 = torch.randn(2, 3, 256, 256).to(device)

    net = scratchai.nets.ENet(4).to(device)

    o1 = net(n1)
    self.assertEqual(list(o1.shape), [2, 4, 256, 256], "out shape reduction not as it should"
                              " be.");
    del n1, o1
    n1 = torch.randn(2, 3, 360, 512).to(device)
    o1 = net(n1)
    self.assertEqual(list(o1.shape), [2, 4, 360, 512], "out shape reduction not as it should"
                              " be.")


class TestResnet(unittest.TestCase):
  
  def test_resnet_init(self):
    noise = torch.randn(2, 3, 224, 224)
    models = ['resnet18', 'resnet34', 'resnet50', 'resnet101', 'resnet152']
    for model in models:
      nc = np.random.randint(1, 1000)
      net = getattr(scratchai.nets, model)(nc=nc, pretrained=False)
      out = net(noise)
      # TODO Assert device is on cpu after initialization
      self.assertEqual(list(out.shape), [2, nc], "out shape not looking good")

      net = getattr(scratchai.nets, model)(pretrained=False)
      out = net(noise)
      self.assertEqual(list(out.shape), [2, 1000], "out shape not looking good")
      del net, out

  def test_resnet18_mnist(self):
    noise = torch.randn(2, 1, 28, 28)
    net = getattr(scratchai.nets, 'resnet18_mnist')()
    out = net(noise)
    # TODO Assert device is on cpu after initialization
    self.assertEqual(list(out.shape), [2, 1000], "out shape not looking good")
    del net, out
    

class TestITN(unittest.TestCase):
  
  n1 = torch.randn(2, 3, 256, 256)
  def test_itnst_(self):
    out = nets.ITN_ST_()(TestITN.n1)
    self.assertEqual(list(out.shape), [2, 3, 256, 256], "out shape not looking good")

  def test_itnst(self):
    out = nets.ITN_ST()(TestITN.n1)
    self.assertEqual(list(out.shape), [2, 3, 256, 256], "out shape not looking good")


class TestLenet(unittest.TestCase):
 
  def test_lenet(self):
    n1 = torch.randn(2, 3, 32, 32)
    out = nets.Lenet(11)(n1)
    # TODO Assert device is on cpu after initialization
    self.assertEqual(list(out.shape), [2, 11], "out shape not looking good")


class TestAlexnet(unittest.TestCase):
  def test_alexnet(self):
    n1 = torch.randn(2, 3, 224, 224)
    net = nets.alexnet(nc=11, pretrained=False); out = net(n1)
    self.assertEqual(list(out.shape), [2, 11], "out shape not looking good")
    out = nets.alexnet(pretrained=False)(n1)
    self.assertEqual(list(out.shape), [2, 1000], "out shape not looking good")

    conv_dict = {0 : [(11, 11), (4, 4), (2, 2)],
                 3 : [(5, 5), (1, 1), (2, 2)],
                 (6, 8, 10) : [(3, 3), (1, 1), (1, 1)]}
    for key, val in conv_dict.items():
      if isinstance(key, tuple):
        for k in key:
          self.assertEqual(net.net[k].kernel_size, val[0], 'not good!')
          self.assertEqual(net.net[k].stride, val[1], 'not good!')
          self.assertEqual(net.net[k].padding, val[2], 'not good!')
      else:
        self.assertEqual(net.net[key].kernel_size, val[0], 'not good!')
        self.assertEqual(net.net[key].stride, val[1], 'not good!')
        self.assertEqual(net.net[key].padding, val[2], 'not good!')
    for i in [1, 4, 7, 9, 11, 17, 20]:
      self.assertIsInstance(net.net[i], nn.ReLU, 'not good!')

  def test_alexnet_mnist(self):
    n1 = torch.randn(2, 1, 28, 28)
    out = nets.alexnet_mnist(pretrained=False)(n1)
    self.assertEqual(list(out.shape), [2, 10], "out shape not looking good")
    

class TestVGG(unittest.TestCase):
  
  def test_vgg(self):
    n1 = torch.randn(2, 3, 224, 224)
    ns = ['vgg{}', 'vgg{}_bn']
    for c in ['11', '13', '16', '19']:
      for n in ns:
        net = getattr(nets, n.format(c))()
        out = net(n1)
        self.assertEqual(list(out.shape), [2, 1000], 'Nope!')
      
class TestCommon(nn.Module):
  
  def test_flatten(self):
    n1 = torch.randn(13, 3, 4, 2)
    out = nets.Flatten()(n1)
    self.assertTrue(out.shape, [13, 3*4*2], 'out shape not okay')

  def test_debug(self):
    utils.implemented(nets, 'debug')


#############################################################
######## Weght Initializations
#############################################################

class WeightInit(unittest.TestCase):

  def test_weight_init(self):
    net = nets.resnet18(pretrained=False)
    # TODO Add more tests
    net.apply(init.xavier_normal)
    net.apply(init.xavier_uniform)
    net.apply(init.kaiming_normal)
    net.apply(init.kaiming_uniform)
    net.apply(init.msr_init)



"""
Image Transformation Network from Justin Johnson et al. 2016. 
"""

import torch
import torch.nn as nn
import torch.nn.functional as F

from torchvision.transforms import CenterCrop

from scratchai.nets.clf.resnet import conv as _conv
from scratchai.nets.seg.enet import uconv as _uconv

def conv(ic:int, oc:int, k:int=3, s:int=2, act:bool=True, 
         norm:bool=True) -> list:
  layers = [nn.ReflectionPad2d(k//2), nn.Conv2d(ic, oc, k, s, bias=True)]
  if norm: layers += [nn.InstanceNorm2d(oc, affine=True)]
  if act: layers += [nn.ReLU(inplace=True)]
  return layers


class uconv(nn.Module):
  """
  The Upsampling Module.

  This module doesn't uses ConvTranspose2d but instead uses nearest mode
  to upsample the image and then passes it through a Conv Layer as it is 
  shown to work better than ConvTranspose2d.
  Link: http://distill.pub/2016/deconv-checkerboard/

  Arguments
  ---------
  ic : int
       # of input channels
  oc : int
       # of output channels
  k : int, tuple
      Kernel_size
  s : int, tuple
      stride
  p : int, tuple
      Padding.
  up : int
       The upsampling factor
  """
  def __init__(self, ic:int, oc:int, k:int=3, s:int=1, act:nn.Module=nn.ReLU, 
               norm:nn.Module=nn.InstanceNorm2d, up:int=2):
    super().__init__()
    self.conv = nn.Sequential(*conv(ic, oc, k, s, act, norm))
    self.up = up
  def forward(self, x): return self.conv(F.interpolate(x, mode='nearest',
                                         scale_factor=self.up))


class resblock(nn.Module):
  """
  This module is the implementation of a resnet block.

  Arguments
  ---------
  ic : int
       # of input channels
  oc : int
       # of output channels
  norm : nn.Module
         The Normalization to be used.
  """

  def __init__(self, ic:int, oc:int=None, norm:nn.Module=nn.InstanceNorm2d):
    super().__init__()
    self.main = nn.Sequential(*conv(ic, oc, s=1), *conv(oc, oc, s=1, act=None))
  def forward(self, x): return self.main(x) +  x


class ITN_ST(nn.Module):
  """
  This module implements the Image Transformation Network proposed in the
  Perceptual Losses paper by Justin et al. 2016 with tweaks.

  Paper Link: 
  https://cs.stanford.edu/people/jcjohns/papers/eccv16/JohnsonECCV16.pdf

  Notes
  -----
  This module is the recommended implementation of the model. Based on
  researches done after the release of the paper.
  """

  def __init__(self):
    super().__init__()
    layers = [*conv(3, 32, 9, 1), *conv(32, 64), *conv(64, 128)]
    layers += [resblock(128, 128, norm=nn.InstanceNorm2d) for _ in range(5)]
    layers += [uconv(128, 64), uconv(64, 32), uconv(32, 3, 9, up=1, norm=None, act=None)]
    self.net = nn.Sequential(*layers)

  def forward(self, x): return self.net(x)    

# Below is the original implementation of the Transformation Network
# as proposed by Justin et al. 2016. Research after the release of that paper
# shows the above architecture to work best. And the above architecture is made
# available by default in scratchai.  The below network can be called using 
# net = scratchai.nets.ITN_ST_()

class _resblock(nn.Module):
  """
  This module is the implementation of a resnet block.
  Arguments
  ---------
  ic : int
       # of input channels
  oc : int
       # of output channels
  norm : nn.Module
         The Normalization to be used.
  """

  def __init__(self, ic:int, oc:int=None, norm:nn.Module=nn.InstanceNorm2d):
    super().__init__()
    self.main = nn.Sequential(*_conv(ic, oc, 3, 1, 0, norm=norm), 
                              *_conv(oc, oc, 3, 1, 0, act=None, norm=norm))
    self.act = nn.ReLU(inplace=True)

  def forward(self, x):
    return self.act(self.main(x) +  x[:, :, 2:-2, 2:-2])

class ITN_ST_(nn.Module):
  """
  This module implements the Image Transformation Network as proposed in the
  Perceptual Losses paper by Justin et al. 2016

  Paper Link: 
  https://cs.stanford.edu/people/jcjohns/papers/eccv16/JohnsonECCV16.pdf

  Notes
  -----
  This is the original implementation of the transformation network.
  """

  def __init__(self):
    super().__init__()

    layers = [nn.ReflectionPad2d(40), *_conv(3, 32, 9, 1, 4), 
              *_conv(32, 64, 3, 2, 1), *_conv(64, 128, 3, 2, 1)]
    layers += [_resblock(128, 128, norm=nn.InstanceNorm2d) for _ in range(5)]
    layers += [*_uconv(128, 64, 4, 2, 1), *_uconv(64, 32, 4, 2, 1), 
               *_uconv(32, 3, 9, 1, 4)]

    self.net = nn.Sequential(*layers)

  def forward(self, x): return self.net(x)



"""
The Noise Attack
"""

import numpy as np
import torch

from torchvision import transforms as T
from torchvision import datasets

__all__ = ['noise', 'Noise']

def noise(x, eps=0.3, order=np.inf, clip_min=None, clip_max=None):
    """
    A weak attack that just picks a random point in the attacker's action
    space. When combined with an attack bundling function, this can be used to
    implement random search.

    References:
    https://arxiv.org/abs/1802.00420 recommends random search to help identify
        gradient masking

    https://openreview.net/forum?id=H1g0piA9tQ recommends using noise as part
        of an attack building recipe combining many different optimizers to
        yield a strong optimizer.

    Arguments
    ---------
    x : torch.Tensor
        The input image.
    """
    
    if order != np.inf: raise NotImplementedError(ord)
    
    eta = torch.FloatTensor(*x.shape).uniform_(-eps, eps).to(x.device)
    adv_x = x + eta

    if clip_min is not None and clip_max is not None:
        adv_x = torch.clamp(adv_x, min=clip_min, max=clip_max)

    return adv_x


###################################################################
# A class to initialize the noise attack
# This class is implemented mainly so this attack can be directly 
# used along with torchvision.transforms

class Noise():
  def __init__(self, net=None, **kwargs):
    self.kwargs = kwargs
  def __call__(self, x):
    return noise(x, **self.kwargs)



"""
The metrics used to measure the performance of models.
"""

import torch
import numpy as np
from tabulate import tabulate


def miou(pred, gt, nc, c2n=None):
  """
  Mean Intersection over Union (mIOU).

  Arguments
  --------
  pred : torch.tensor, [N x 3 x H x W]
        The original input images to the model.
  gt : torch.tensor, [N x H x W]
        The corresponding labels of the images.
  nc : int
       The number of classes
  c2n : dict
        The mapping from class idx to class names.

  Returns
  -------
  miou : float
         The mean intersection over union value.
  
  Notes
  -----
  Do note: if batches of data are passed it is necessary that
  inp - [N x H x W]
  lab - [N x H x W]

  where each matrix in lab is have each pixel value in the range of [0, C)
  where C is the number of classes.
  """
  # TODO This function needs testing

  assert len(list(pred.shape)) in [3, 2]
  # Assert batch_size, height and width matches
  assert pred.shape == gt.shape
  
  with torch.no_grad():
    # Convert torch.tensor to np.ndarray
    if isinstance(pred, torch.Tensor):
      pred = pred.clone().detach().cpu().numpy()
    if isinstance(gt, torch.Tensor):
      gt = gt.clone().detach().cpu().numpy()

    iou = {}
    miou = 0
    for cls in range(nc):
      inter = np.logical_and(pred == cls,  gt == cls).sum()
      union = np.logical_or(pred == cls, gt == cls).sum()
      iou[cls] = inter / union if inter != 0 and union != 0 else 1.
      miou += iou[cls]
  
  # TODO print the iou in table format
  #print (tabulate([list(iou.keys()), list(iou.values())]))

  return miou/nc


def accuracy(out:torch.Tensor, target:torch.Tensor, topk:tuple=(1,)) -> list:
  """
  Function to help in measuring the accuracy of the predicted labels
  against the true labels.

  Arguments
  ---------
  out : torch.Tensor, [N x C] where C is the number of classes
        The predicted logits 
  pred : torch.Tensor, [N,]
          Total number of elements.
  """ 
  
  with torch.no_grad():
    assert out.shape[0] == target.shape[0]
    assert max(target) <= out.shape[1]
    _, pred = out.topk(max(topk), 1, True, True); pred.t_()
    corr = pred.eq(target.view(1, -1).expand_as(pred))
    acc_list = []
    for k in topk:
      corr_k = corr[:k].sum().item()
      acc_list.append(corr_k / out.size(0))
    return acc_list



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="scratchai-nightly",
    version="0.0.1a2",
    author="@iArunava",
    author_email="iarunavaofficial@gmail.com",
    description="Scratch AI",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/iArunava/scratch.ai",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



"""
Alexnet

https://papers.nips.cc/paper/4824-imagenet-classification-with-deep-
convolutional-neural-networks.pdf
from Alex Krizhevsky Ilya Sutskever Geoffrey E. Hinton
"""

import torch
import torch.nn as nn

from scratchai.nets.common import Flatten
from scratchai.utils import load_pretrained
from scratchai.pretrained import urls


__all__ = ['alexnet', 'alexnet_mnist']


def conv(ic:int, oc:int, k:int=3, s:int=1, p:int=1, pool:bool=True):
  layers = [nn.Conv2d(ic, oc, k, s, p), nn.ReLU(inplace=True)]
  if pool: layers += [nn.MaxPool2d(3, 2)]
  return layers

def linear(inn:int, otn:int):
  return [nn.Dropout(), nn.Linear(inn, otn), nn.ReLU(inplace=True)]

class Alexnet(nn.Module):
  """
  Implmentation of Alexnet.  
  
  Arguments 
  --------- 
  nc : int 
       # of classes 
  ic : int 
       # of channels

  References
  ----------
  https://papers.nips.cc/paper/4824-imagenet-classification-with-deep-
  convolutional-neural-networks.pdf
  """
  def __init__(self, nc:int=1000, ic:int=3):
    super().__init__()
    # Special Case: MNIST.
    ic2 = 64 if ic == 3 else 1
    layers = [*conv(ic, 64, 11, 4, 2), *conv(ic2, 192, 5, p=2), 
              *conv(192, 384, pool=False), *conv(384, 256, pool=False), 
              *conv(256, 256), nn.AdaptiveAvgPool2d((6, 6)), 
              Flatten(), *linear(256*6*6, 4096), *linear(4096, 4096), 
              nn.Linear(4096, nc)]
    # Special Case: MNIST. Removing the first conv->relu->pool layer
    if ic == 1: layers.pop(0); layers.pop(0); layers.pop(0)
    self.net = nn.Sequential(*layers)
    
  def forward(self, x): return self.net(x)


def alexnet_mnist(pretrained=True, **kwargs):
  cust_nc = kwargs['nc'] if 'nc' in kwargs else None
  kwargs['ic'] = 1; kwargs['nc'] = 10
  net = Alexnet(**kwargs)
  if pretrained:
    return load_pretrained(net, urls.alexnet_mnist_url, 'alexnet_mnist', 
                           nc=cust_nc, attr='net[21]', inn=4096)
  return net

def alexnet(pretrained=True, **kwargs):
  cust_nc = None
  if pretrained and 'nc' in kwargs: cust_nc = kwargs['nc']; kwargs['nc'] = 1000
  net = Alexnet(**kwargs)
  if pretrained:
    return load_pretrained(net, urls.alexnet_url, 'alexnet', nc=cust_nc, 
                           attr='net[21]', inn=4096)
  return net



"""
The Fast Gradient Method attack.
"""

import numpy as np
import torch
import torch.nn as nn

from scratchai.attacks.utils import optimize_linear


__all__ = ['fgm', 'FGM']

def fgm(x, net:nn.Module, eps:float=0.3, ordr=np.inf, y=None, 
        clip_min=None, clip_max=None, targeted=False, sanity_checks=True):
  """
  Implementation of the Fast Gradient Method.

  Arguments
  ---------
  net : nn.Module
        The model on which to perform the attack.
  x : torch.tensor
      The input to the net.
  y : torch.tensor, optional
      True labels. If targeted is true, then provide the target label.
      Otherwise, only model predictions are used as labels to avoid 
      the "label leaking" effect (explained in this paper: 
      https://arxiv.org/abs/1611.01236). Default is None.
      Labels should be one hot encoded.
  eps: float, optional
       The epsilon (input variation parameter)
  ord : [np.inf, 1, 2], optional
        Order of the norm.
  clip_min : float
             Minimum float value for adversarial example components.
  clip_max : float
             Maximum float value for adversarial example components.
  targeted : bool, optional
             Is the attack targeted or untargeted? Untargeted will try to make
             the label incorrect. Targeted will instead try to move in the
             direction of being more like y.

  Returns
  -------
  adv_x : torch.Tensor
          The adversarial example.
  """
  if ordr not in [np.inf, 1, 2]:
    raise ValueError('Norm order must be either np.inf, 1, or 2.')

  if clip_min:
    assert torch.all(x > torch.tensor(clip_min, device=x.device, dtype=x.dtype))
  if clip_max:
    assert torch.all(x < torch.tensor(clip_max, device=x.device, dtype=x.dtype))
  
  # Flag to indicate if the image has been unsqueezed
  usq = False
  # Inplace operations not working for some bug #15070
  # TODO Update when fixed
  if len(x.shape) == 3: x = x.unsqueeze(0); usq = True

  # x needs to have requires_grad set to True 
  # for its grad to be computed and stored properly in a backward call
  x = x.detach().clone(); x.requires_grad_(True)

  if y is None: _, y = torch.max(net(x), dim=1)

  # Compute loss
  crit = nn.CrossEntropyLoss()
  loss = crit(net(x), y)
  # If attack is targeted, minimize loss of target label rather than maximize
  # loss of correct label.
  if targeted: loss = -loss

  # Define gradient of loss wrt input
  loss.backward()
  opt_pert = optimize_linear(x.grad, eps, ordr)

  # Add perturbation to original example to obtain adversarial example
  adv_x = x + opt_pert

  # If clipping is needed, reset all values outside of [clip_min, clip_max]
  if (clip_min is not None) or (clip_max is not None):
    # We don't currently support one-sided clipping
    assert clip_min is not None and clip_max is not None
    adv_x = torch.clamp(adv_x, clip_min, clip_max)
  
  return adv_x if not usq else adv_x.squeeze()

###################################################################
# A class to initialize the attack
# This class is implemented mainly so this attack can be directly
# used along with torchvision.transforms

class FGM():
  def __init__(self, net, **kwargs):
    self.net = net
    self.kwargs = kwargs
  def __call__(self, x):
    return fgm(x, self.net, **self.kwargs)



from scratchai.attacks.attacks.noise import *
from scratchai.attacks.attacks.semantic import *
from scratchai.attacks.attacks.fast_gradient_method import *
from scratchai.attacks.attacks.pgd import *
from scratchai.attacks.attacks.deepfool import *



"""
DeepFool Attack from https://arxiv.org/pdf/1511.04599.pdf 
by Seyed Mohsen et al.
"""

import torch
import torch.nn as nn
import numpy as np
import copy

from torch.autograd.gradcheck import zero_gradients as zgrad


__all__ = ['deepfool', 'DeepFool']


def deepfool(x, net:nn.Module, eta:float=0.02, tnc:int=10, miter:int=50, 
             mret=False):
  """
  Implementation of the DeepFool Attack.

  Arguments
  ---------
  x : torch.Tensor
      The input tensor
  net : nn.Module
        The net.
  eta : float
        Eta as used in the paper.
  tnc : int
       The number of classes to consider.
  miter : int
          The maximum number of iterations.
  mret : bool
         If True returns the minimal perturbation that was added 
         to the original image. Defaults to False.
  """
  # Set net to eval
  net.eval()
  dev = x.device

  # If x doesn't have a batch dimension add 1
  # The loop runs twice to ensure that 2D images gets the batch dim
  # For 2D images, 1st iter adds -> The channel dim
  # 2nd iter adds -> batch dim
  for _ in range(2): 
    if len(x.shape) < 3: x = x.unsqueeze(0)
  bs = x.shape[0]
  lbs = list(range(bs))
  
  # x_idt is the initial advesarial image
  x_idt = x.detach().clone().requires_grad_(True)
  # Get the logits (x_idt is not perturbed yet, the logits are true logits)
  logits = net(x_idt)
  # Get the top tnc classes that the net is most confident about
  psort = torch.argsort(logits.data.cpu(), dim=1, descending=True)
  psort = psort[:, :tnc].data.numpy()
  
  tlabl = psort[:, 0] # True Label
  plabl = tlabl       # Pert label

  x_shape = x_idt.shape  # [N x C x H x W]
  w = np.zeros(x_shape)  # [N x C x H x W]
  rt = np.zeros(x_shape) # [N x C x H x W]
  
  i = 0
  while (plabl == tlabl).any() and i < miter:
    # Initial Perturbation
    pert = 5e10
    logits[lbs, tlabl].sum().backward(retain_graph=True)
    ograd = x_idt.grad.data.cpu().numpy().copy()

    for c in range(1, tnc):
      zgrad(x_idt)
      logits[lbs, psort[:, c]].sum().backward(retain_graph=True)
      cgrad = x_idt.grad.data.cpu().numpy().copy()

      # Get new wk and fk
      wk = cgrad - ograd
      fk = (logits[lbs, psort[:, c]] - logits[lbs, tlabl]) \
             .detach().cpu().data.numpy()

      cpert = abs(fk) / np.linalg.norm(wk.reshape(bs, -1), axis=1)
      blist = (cpert < pert).astype(np.float32)
      if blist.any(): 
        pert = pert*(1-blist) + (cpert*blist)
        w = (w.T*(1-blist) + (wk.T * blist)).T
    
    # Added 1e-4 for numerical stability
    ri =  (((pert+1e-4) * w.T) / np.linalg.norm(w.reshape(bs, -1), axis=1)).T
    rt += ri
    
    x_idt = x + ((1+eta) * torch.from_numpy(rt)).float().to(dev)
    x_idt.requires_grad_(True); logits = net(x_idt)
    plabl = torch.argmax(logits, dim=1).detach().cpu().data.numpy()

    i += 1
  
  rt = (1+eta) * rt
  return (x_idt, rt) if mret else x_idt

##################################################################

class DeepFool():
  def __init__(self, net, **kwargs):
    self.net = net
    self.kwargs = kwargs
  def __call__(self, x):
    return deepfool(x, self.net, **self.kwargs)



import unittest
import torch
import torch.nn as nn
from scratchai import *

class TestLearners(unittest.TestCase):
    
  def test_clf_test(self):
    # Check if the function is implemented
    utils.implemented(learners.clflearner, 'clf_test')

  def test_clf_train(self):
    # Check if the function is implemented
    utils.implemented(learners.clflearner, 'clf_train')

  def test_clf_fit(self):
    # Check if the function is implemented
    utils.implemented(learners.clflearner, 'clf_fit')

  def test_imp_mnist_cifar10(self):
    # Check if the function is implemented
    utils.implemented(learners.quicktrain, 'mnist')
    utils.implemented(learners.quicktrain, 'cifar10')
  
  @unittest.skipIf(torch.cuda.is_available() == False, 'no cuda')
  def test_train_mnist_cifar10(self):
    # Check if the function is implemented
    net = nets.lenet_mnist()
    learners.quicktrain.mnist(net, epochs=1)
    net = nets.lenet_cifar10()
    learners.quicktrain.cifar10(net, epochs=1)

  def test_adjust_lr(self):
    # Check if the function is implemented
    utils.implemented(learners.clflearner, 'adjust_lr')
